//-------------------------------------------------------------------------
// 
// Copyright Institut Telecom, ETRI and Hanyang University 2008, 2009
// Licensed Materials, Program Property of Institut Telecom, ETRI and Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  Institut Telecom 
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Khaled Mammou
//    Marius Preda
//    Seung Wook Lee
//    Daiyong Kim
//    KyoungSoo Son
//
//
//    http://www-artemis.it-sudparis.eu/
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//    
//  
//    Marius Preda       (marius.preda@int-evry.fr)
//    Khaled Mammou      (khaled_mamou@yahoo.fr)
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------
#pragma warning (disable : 4786)

#include <String.h>
#include <math.h>
#include <time.h>
//#include <iostream>
#include "SC3DMC_IndexedFaceSet.h"
#include "SC3DMC_SC3DMCEncoder.h"
#include "SC3DMC_EncodingParameter.h"

bool Configure(char* filename, EncodingParams* par);

int main ( int argc, char *argv[])
{

	printf("Usage:\n");
	printf("SC3DMCEncoder inputFileName paramFileName\n\n");

	printf("	inputFileName:   a VRML 2.0 describing a 3D triangular mesh \n");
	printf("	paramFileName:   file describing the encoding parameters\n");

	if (argc != 3) {
		printf("Error: we need 2 arguments! (see usage)\n");
		return -1;
	}
	/*	
	printf("This software was developped by: \n");
	printf(" - Khaled Mammou (khaled_mamou@yahoo.fr, Institut Telecom)\n");
	printf(" - Seung Wook Lee (tajinet@etri.re.kr, ETRI/Hanyang University)\n");	
	printf(" - Daiyong Kim (dykim@dmlab.hanyang.ac.kr, ETRI/Hanyang University)\n");
	printf(" - KyoungSoo Son (bonokensin@dmlab.hanyang.ac.kr, ETRI/Hanyang University)\n");
	printf(" - Marius Preda (marius.preda@int-evry.fr, Institut Telecom)\n\n");

	printf("Copyright Institut Telecom (http://www-artemis.it-sudparis.eu/), ETRI (http://www.etri.re.kr/) and Hanyang University (http://dmlab.hanyang.ac.kr/) 2008, 2009\n");
	printf("Licensed Materials, Program Property of Institut Telecom, ETRI and Hanyang University.\n\n\n");
*/
	char configfilename[50];
	strcpy(configfilename,argv[2]);

	char fin[1024];
	char filename[100];
	strcpy(filename,argv[1]);
	sprintf(fin, "%s.wrl", filename);

	char fout[1024];
	sprintf(fout, "%s.s3d", filename);
//
	printf("inputFileName\t\t%s\n",fin); 
	printf("paramFileName\t\t%s\n",configfilename); 
//


	EncodingParams params;
	Configure(configfilename,&params);
//
	printf("encodingMode\t\t%i\n", params.m_encodingMode);
	printf("binarizationMode\t\t%i\n", params.m_binarizationMode);
	printf("qCoord\t\t%i\n", params.m_nQCoordBits);
	printf("qTexture\t\t%i\n", params.m_nQTexCoordBits);
	printf("qNormal\t\t%i\n", params.m_nQNormalBits);
	printf("qColor\t\t%i\n", params.m_nQColorBits);
	printf("qOtherAttributes\t\t%i\n", params.m_nQOtherAttributesBits);

	printf("coordPred\t\t%i\n", params.m_coordPredMode);
	printf("texturePred\t\t%i\n", params.m_texCoordPredMode);
	printf("normalPred\t\t%i\n", params.m_normalPredMode);
	printf("colorPred\t\t%i\n", params.m_colorPredMode);
	printf("otherAttributesPred\t\t%i\n", params.m_otherAttributesPredMode);

	printf("vertexOrderPres\t\t%i\n", params.m_vertexOrderPres);
	printf("triangleOrderPres\t\t%i\n", params.m_triangleOrderPres);
//



	SC3DMC_IndexedFaceSet myIFS;
	// read file
	myIFS.LoadIFSVRML2(fin);
	
	double elapsedTime = 0.0;
	clock_t stopTime = 0;
	clock_t startTime = clock();

	SC3DMCEncoder myEncoder;
	myEncoder.SetIFS(&myIFS);
	myEncoder.SetParams(params);
	unsigned char * compressedStream = new unsigned char[myEncoder.GetEstimatedCompressedStreamSize()];
	unsigned int compressedStreamSize = 0;
	myEncoder.Encode(compressedStream, compressedStreamSize);
	WriteStream2File(fout, compressedStream, compressedStreamSize);
	delete [] compressedStream;


	stopTime = clock();
	elapsedTime = (stopTime - startTime) / (double)CLOCKS_PER_SEC;
	printf("%s\t%d\t%f\n", filename, compressedStreamSize, elapsedTime);
	return 0;
}

//dmlab
bool Configure(char* filename, EncodingParams* par)
{
	FILE *pfPara;

	if ((pfPara = fopen (filename, "r")) == NULL ){
		printf ("Parameter File Not Found\n");
		exit (1);
	}


	char rgBuf[80];
	char* tgBuf= NULL;
	char *pch = fgets(rgBuf, 79, pfPara);

	if(strncmp("!! SC3DMC !!", rgBuf, 12)!=0)
	{
		printf("Bad magic number at start of parameter file\n");
		exit(1);
	}

	while((pch = fgets(rgBuf, 79, pfPara)) != NULL)
	{
		//	printf("%s",rgBuf);
		if((strstr(rgBuf,"EncodingMode")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_encodingMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"BinarizationMode")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_binarizationMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"QuantizationParameter.Coord")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_nQCoordBits = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"QuantizationParameter.Normal")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_nQNormalBits = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"QuantizationParameter.Color")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_nQColorBits = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"QuantizationParameter.TexCoord")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_nQTexCoordBits = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"QuantizationParameter.OtherAttributes")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_nQOtherAttributesBits = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"TexCoord.Height")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_nQTexCoordHeight = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"TexCoord.Width")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_nQTexCoordWidth = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.Coord")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_coordPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.Normal")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_normalPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.Color")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_colorPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.TexCoord")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_texCoordPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.OtherAttribute")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_otherAttributesPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.IndexCoord")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_coordIndexPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.IndexNormal")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_normalIndexPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.IndexColor")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_colorIndexPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.IndexTexCoord")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_texCoordIndexPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.IndexOtherAttribute")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_otherAttributesIndexPredMode = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.VertexOrder")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_vertexOrderPres = atoi(tgBuf);
		}
		else if((strstr(rgBuf,"PredictionMode.TriangleOrder")) != NULL)
		{
			tgBuf = strtok(rgBuf,"=");
			tgBuf = strtok(NULL," \n");
			par->m_triangleOrderPres = atoi(tgBuf);
		}
	}


	fclose(pfPara);
	return true;
}
//~dmlab